@@ -11,7 +11,6 @@ from os import getenv
 from pathlib import Path, PurePosixPath
 from platform import system
 from shutil import which
-from socket import inet_aton
 from subprocess import PIPE, STDOUT, TimeoutExpired, check_output, run
 from tempfile import TemporaryDirectory
 from time import sleep, time
@@ -59,6 +58,8 @@ def _get_android_sdk() -> Path:
     return Path.home() / "Android" / "Sdk"
 
 
+AAPT_BIN: str | None = None
+ADB_BIN: str | None = None
 ANDROID_SDK_ROOT = _get_android_sdk()
 DEVICE_TMP = PurePosixPath("/data/local/tmp")
 
@@ -80,10 +81,7 @@ class ADBSession:
     """ADB Session management.
 
     Attributes:
-        _adb_bin: ADB binary to use.
         _debug_adb: Include ADB output in debug output.
-        _ip_addr: Target device IP address.
-        _port: ADB listening port.
         _root: Connected as root user.
         connected: ADB connection state.
         device_id: Unique Android ID.
@@ -91,40 +89,24 @@ class ADBSession:
     """
 
     __slots__ = (
-        "_adb_bin",
         "_debug_adb",
-        "_ip_addr",
-        "_port",
         "_root",
+        "_serial",
         "connected",
         "device_id",
         "symbols",
     )
 
-    def __init__(self, ip_addr: str | None = None, port: int = 5555) -> None:
-        self._adb_bin = self._adb_check()
+    def __init__(self, serial: str) -> None:
         self._debug_adb = getenv("SHOW_ADB_DEBUG", "0") != "0"
-        self._ip_addr: str | None = None
-        self._port: int | None = None
         self._root = False
+        self._serial = serial
         self.connected = False
         self.device_id: str | None = None
         self.symbols: dict[str, Path] = {}
 
-        if ip_addr is not None:
-            LOG.debug("creating IP based session")
-            try:
-                if ip_addr != "localhost":
-                    inet_aton(ip_addr)
-            except (OSError, TypeError):
-                raise ValueError("Invalid IP Address") from None
-            self._ip_addr = ip_addr
-            if not 0x10000 > port > 1024:
-                raise ValueError("Port must be valid integer between 1025 and 65535")
-            self._port = port
-
-    @classmethod
-    def _aapt_check(cls) -> Path:
+    @staticmethod
+    def _aapt_check() -> str:
         """Find Android Asset Packaging Tool (AAPT).
         An OSError is raised if the AAPT executable is not found.
 
@@ -134,19 +116,24 @@ class ADBSession:
         Returns:
             AAPT binary.
         """
-        skd_bin = ANDROID_SDK_ROOT / "android-9" / "aapt"
-        if skd_bin.is_file():
-            LOG.debug("using recommended aapt from '%s'", skd_bin)
-            return skd_bin
-        installed_bin = which("aapt")
-        if installed_bin is None:
-            raise OSError("Please install AAPT")
-        # TODO: update this to check aapt version
-        LOG.warning("Using aapt binary from '%s'", installed_bin)
-        return Path(installed_bin)
+        # pylint: disable=global-statement
+        global AAPT_BIN
+        if AAPT_BIN is None:
+            skd_bin = ANDROID_SDK_ROOT / "android-9" / "aapt"
+            if skd_bin.is_file():
+                LOG.debug("using recommended aapt from '%s'", skd_bin)
+                AAPT_BIN = str(skd_bin)
+            else:
+                installed_bin = which("aapt")
+                if installed_bin is None:
+                    raise OSError("Please install AAPT")
+                # TODO: update this to check aapt version
+                LOG.warning("Using aapt binary from '%s'", installed_bin)
+                AAPT_BIN = installed_bin
+        return AAPT_BIN
 
-    @classmethod
-    def _adb_check(cls) -> Path:
+    @staticmethod
+    def _adb_check() -> str:
         """Find Android Debug Bridge (ADB).
         An OSError is raised if the ADB executable is not found.
 
@@ -156,22 +143,28 @@ class ADBSession:
         Returns:
             ADB binary.
         """
-        sdk_bin = ANDROID_SDK_ROOT / "platform-tools" / "adb"
-        if sdk_bin.is_file():
-            LOG.debug("using recommended adb from '%s'", sdk_bin)
-            return sdk_bin
-        installed_bin = which("adb")
-        if installed_bin is None:
-            raise OSError("Please install ADB")
-        # TODO: update this to check adb version
-        LOG.warning("Using adb binary '%s'", installed_bin)
-        LOG.warning("You are not using the recommended ADB install!")
-        LOG.warning("Either run the setup script or proceed with caution.")
-        sleep(5)
-        return Path(installed_bin)
+        # pylint: disable=global-statement
+        global ADB_BIN
+        if ADB_BIN is None:
+            sdk_bin = ANDROID_SDK_ROOT / "platform-tools" / "adb"
+            if sdk_bin.is_file():
+                LOG.debug("using recommended adb from '%s'", sdk_bin)
+                ADB_BIN = str(sdk_bin)
+            else:
+                installed_bin = which("adb")
+                if installed_bin is None:
+                    raise OSError("Please install ADB")
+                # TODO: update this to check adb version
+                LOG.warning("Using adb binary from '%s'", installed_bin)
+                LOG.warning("You are not using the recommended ADB install!")
+                LOG.warning("Either run the setup script or proceed with caution.")
+                ADB_BIN = installed_bin
+        return ADB_BIN
 
-    @staticmethod
-    def _call_adb(cmd: Sequence[str], timeout: int) -> ADBResult:
+    @classmethod
+    def _call_adb(
+        cls, cmd: Sequence[str], timeout: int = 10, serial: str | None = None
+    ) -> ADBResult:
         """Wrapper to make calls to ADB. Launches ADB in a subprocess and collects
         output. If timeout is specified and elapses the ADB subprocess is terminated.
         This function is only meant to be called directly by ADBSession.call().
@@ -179,13 +172,15 @@ class ADBSession:
         Args:
             cmd: Full ADB command.
             timeout: Seconds to wait for ADB command to complete.
+            serial: Serial of Android device to connect to.
 
         Returns:
             Exit code and stderr, stdout of ADB call.
         """
+        adb = cls._adb_check()
         try:
             result = run(
-                cmd,
+                (adb, "-s", serial, *cmd) if serial else (adb, *cmd),
                 check=False,
                 encoding="utf-8",
                 errors="replace",
@@ -282,7 +277,7 @@ class ADBSession:
         # a few adb commands do not require a connection
         if not self.connected and args[0] not in {"connect", "devices", "disconnect"}:
             raise ADBCommunicationError("ADB session is not connected!")
-        result = self._call_adb((str(self._adb_bin), *args), timeout=timeout)
+        result = self._call_adb(args, timeout=timeout, serial=self._serial)
         if self._debug_adb:
             LOG.debug(
                 "=== adb start ===\n%s\n=== adb end, returned %d ===",
@@ -339,45 +334,22 @@ class ADBSession:
             cmd.append(f"--pid={pid}")
         return self.call(cmd, timeout=30).output
 
-    def connect(
-        self,
-        as_root: bool = True,
-        boot_timeout: int = 300,
-        max_attempts: int = 60,
-        retry_delay: int = 1,
-    ) -> bool:
+    def connect(self, as_root: bool = True, boot_timeout: int = 300) -> None:
         """Connect to a device via ADB.
 
         Args:
             as_root: Attempt to enable root. Default is True.
             boot_timeout: Seconds to wait for device boot to complete.
-            max_attempts: Number of attempt to connect to the device.
-            retry_delay: Seconds to wait between connection attempts.
 
         Returns:
-            True if connection was established otherwise False.
+            None.
         """
         assert boot_timeout > 0
-        assert max_attempts > 0
-        assert retry_delay >= 0
-        attempt = 0
+        retry = True
         root_called = False
         set_enforce_called = False
-        while attempt < max_attempts:
-            attempt += 1
-            # attempt to connect to device
-            self.connected = False
-            if self._ip_addr is not None:
-                addr = f"{self._ip_addr}:{self._port}"
-                LOG.debug("connecting to '%s'", addr)
-                if self.call(["connect", addr], timeout=30).exit_code != 0:
-                    LOG.warning("connection attempt #%d failed", attempt)
-                    sleep(retry_delay)
-                    continue
-            elif not self.devices():
-                LOG.warning("No device detected (attempt %d/%d)", attempt, max_attempts)
-                sleep(retry_delay)
-                continue
+        while retry:
+            retry = False
             self.connected = True
             # verify we are connected
             LOG.debug("waiting for device to boot (%ds)...", boot_timeout)
@@ -391,84 +363,48 @@ class ADBSession:
                 ["settings", "get", "secure", "android_id"], device_required=False
             )
             if result.exit_code != 0:
-                LOG.error("Failed to retrieve Android ID")
+                LOG.debug("failed to retrieve Android ID")
+                self.connected = False
                 raise ADBSessionError("Device in invalid state")
             self.device_id = result.output
             # get active user
             result = self.shell(["whoami"], device_required=False, timeout=30)
             if result.exit_code != 0 or not result.output:
-                LOG.error("Failed to retrieve active user")
+                LOG.debug("failed to retrieve active user")
                 raise ADBSessionError("Device in invalid state")
             self._root = result.output == "root"
-            # check SELinux mode
-            if self._root and self.get_enforce():
+            if as_root:
+                # handle root login
+                if not self._root:
+                    if root_called:
+                        LOG.debug("root login attempt failed")
+                        raise ADBSessionError("Root login failed")
+                    self.call(["root"], timeout=30)
+                    retry = True
+                    root_called = True
+                    self.connected = False
+                    continue
+                # handle SELinux
                 if set_enforce_called:
                     self.connected = False
                     raise ADBSessionError("set_enforce(0) failed!")
-                # set SELinux to run in permissive mode
-                self.set_enforce(0)
-                self.shell(["stop"])
-                self.shell(["start"])
-                # put the device in a known state
-                self.call(["reconnect"], timeout=30)
-                set_enforce_called = True
-                attempt -= 1
-                continue
-            # enable root if needed
-            if as_root and not self._root:
-                if self.call(["root"], timeout=30).exit_code == 0:
-                    # only skip attempt to call root once
-                    if not root_called:
-                        root_called = True
-                        attempt -= 1
-                else:
-                    LOG.warning("Failed root login attempt")
-                continue
-            # connected!
-            break
-        else:
-            LOG.debug("failed to connect to device")
-            self.connected = False
-            return False
-
+                if self.get_enforce():
+                    # set SELinux to run in permissive mode
+                    self.set_enforce(0)
+                    self.shell(["stop"])
+                    self.shell(["start"])
+                    # put the device in a known state
+                    self.call(["reconnect"], timeout=30)
+                    retry = True
+                    set_enforce_called = True
+                    self.connected = False
+                    continue
         assert self.connected
-        return True
 
-    @classmethod
-    def create(
-        cls,
-        ip_addr: str | None = None,
-        port: int = 5555,
-        as_root: bool = True,
-        max_attempts: int = 10,
-        retry_delay: int = 1,
-    ) -> ADBSession | None:
-        """Create a ADBSession and connect to a device via ADB.
-
-        Args:
-            ip_addr: IP address of device to connect to if using TCP/IP.
-            port: Port to use (TCP/IP only).
-            as_root: Attempt to enable root.
-            max_attempts: Number of attempts to connect to the device.
-            retry_delay: Number of seconds to wait between attempts.
-
-        Returns:
-            A connected ADBSession object or None
-        """
-        session = cls(ip_addr, port)
-        if session.connect(
-            as_root=as_root, max_attempts=max_attempts, retry_delay=retry_delay
-        ):
-            return session
-        return None
-
-    def devices(
-        self, all_devices: bool = False, any_state: bool = True
-    ) -> dict[str, str]:
+    def devices(self, any_state: bool = True) -> dict[str, str]:
         """Devices visible to ADB.
 
         Args:
-            all_devices: Don't filter devices using ANDROID_SERIAL environment variable.
             any_state: Include devices in a state other than "device".
 
         Returns:
@@ -478,22 +414,15 @@ class ADBSession:
         devices: dict[str, str] = {}
         if result.exit_code != 0:
             return devices
-        target_device = getenv("ANDROID_SERIAL", None) if not all_devices else None
         # skip header on the first line
         for entry in result.output.splitlines()[1:]:
             try:
                 name, state = entry.split()
             except ValueError:
                 continue
-            if target_device is not None and name != target_device:
-                continue
             if not any_state and state != "device":
                 continue
             devices[name] = state
-        if target_device is None and not all_devices and len(devices) > 1:
-            raise ADBSessionError(
-                "Multiple devices available and ANDROID_SERIAL not set"
-            )
         return devices
 
     def disconnect(self, unroot: bool = True) -> None:
@@ -511,14 +440,11 @@ class ADBSession:
         if self._root and unroot:
             try:
                 if self.call(["unroot"], timeout=30).exit_code == 0:
-                    self.connected = False
                     self._root = False
-                    return
-                LOG.warning("'unroot' failed")
             except ADBCommandError:
                 LOG.warning("'unroot' not support by ADB")
-        elif self._ip_addr is not None:
-            self.call(["disconnect", f"{self._ip_addr}:{self._port}"], timeout=30)
+            if self._root:
+                LOG.debug("'unroot' failed during disconnect")
         self.connected = False
 
     def get_enforce(self) -> bool:
@@ -761,15 +687,11 @@ class ADBSession:
             raise FileNotFoundError(f"'{path}' does not exist")
         return PurePosixPath(result.output)
 
-    def reboot_device(
-        self, boot_timeout: int = 300, max_attempts: int = 60, retry_delay: int = 1
-    ) -> None:
+    def reboot_device(self, boot_timeout: int = 300) -> None:
         """Reboot the connected device and reconnect.
 
         Args:
             boot_timeout: Seconds to wait for device boot to complete.
-            max_attempts: Number of attempts to connect to the device.
-            retry_delay: Seconds to wait between connection attempts.
 
         Returns:
             None
@@ -777,12 +699,8 @@ class ADBSession:
         was_root = self._root
         self.call(["reboot"])
         self.connected = False
-        self.connect(
-            as_root=was_root,
-            boot_timeout=boot_timeout,
-            max_attempts=max_attempts,
-            retry_delay=retry_delay,
-        )
+        self.connect(as_root=was_root, boot_timeout=boot_timeout)
+        # TODO: better error handling
         assert self.connected, "Device did not connect after reboot"
 
     def remount(self) -> None:
