@@ -353,17 +353,13 @@ class AndroidSDKRepo:
 
 
 def _is_free(port: int) -> bool:
-    sock = socket(AF_INET, SOCK_STREAM)
-    try:
+    with suppress(OSError), socket(AF_INET, SOCK_STREAM) as sock:
         sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
         sock.settimeout(0.05)
         sock.bind(("localhost", port))
         sock.listen()
         return True
-    except OSError:
-        return False
-    finally:
-        sock.close()
+    return False
 
 
 class AndroidEmulatorError(Exception):
@@ -399,6 +395,7 @@ class AndroidEmulator:
         target: str | None = None,
         verbose: bool = False,
         boot_timeout: int = 300,
+        emulator_output: bool = False,
     ) -> None:
         """Create an AndroidEmulator object.
 
@@ -413,9 +410,9 @@ class AndroidEmulator:
             target: The target name (from builds.json).
             verbose: Enable verbose logging.
             boot_timeout: Time to wait for Android to boot in the emulator.
+            emulator_output: Display output from the emulator console.
         """
         self.avd_name = avd_name
-        self.emu = None
         self.env = dict(env or {})
         self.port = port
         self.snapshot = Snapshot[snapshot.upper()]
@@ -460,8 +457,6 @@ class AndroidEmulator:
         args.extend(("-port", f"{self.port:d}"))
         args.append(f"@{self.avd_name}")
 
-        output = None if getLogger().getEffectiveLevel() == DEBUG else DEVNULL
-
         if xvfb:
             try:
                 self.xvfb: Xvfb | None = Xvfb(width=1280, height=1024, timeout=60)
@@ -481,29 +476,28 @@ class AndroidEmulator:
                 env["XAUTHORITY"] = getenv("XAUTHORITY", "")
         env["ANDROID_AVD_HOME"] = str(PATHS.avd_home)
 
-        LOG.info("Launching Android emulator with snapshot=%s", self.snapshot.name)
+        LOG.debug("launching emulator (port=%d, snapshot=%s)", port, self.snapshot.name)
         emu = Popen(  # pylint: disable=consider-using-with
             [str(PATHS.sdk_root / "emulator" / f"emulator{EXE_SUFFIX}"), *args],
             env=env,
-            stderr=output,
-            stdout=output,
+            stderr=None if emulator_output else DEVNULL,
+            stdout=None if emulator_output else DEVNULL,
         )
-
         try:
             self.boot_wait(emu, port, boot_timeout)
-        except Exception as exc:
+        except:
             if emu.poll() is None:
                 emu.terminate()
                 try:
                     # this should not hang (but does), timeout is here just in case.
-                    emu.wait(120)
+                    emu.wait(60)
                 except TimeoutExpired:
                     emu.kill()
                     with suppress(TimeoutExpired):
                         emu.wait(10)
             if self.xvfb is not None:
                 self.xvfb.stop()
-            raise exc
+            raise
 
         self.emu = emu
         self.pid = emu.pid
@@ -523,6 +517,8 @@ class AndroidEmulator:
         assert boot_timeout > 0
         cmd = (
             str(PATHS.sdk_root / "platform-tools" / f"adb{EXE_SUFFIX}"),
+            "-s",
+            f"emulator-{port:d}",
             "shell",
             "getprop",
             "sys.boot_completed",
@@ -530,16 +526,16 @@ class AndroidEmulator:
         LOG.debug("waiting for device to boot...")
         deadline = perf_counter() + boot_timeout
         while True:
-            adb_result = run(
-                cmd,
-                env={"ANDROID_SERIAL": f"emulator-{port:d}"},
-                capture_output=True,
-                check=False,
-                timeout=30,
-            )
-            if adb_result.returncode == 0 and adb_result.stdout.strip() == b"1":
-                LOG.debug("device booted")
-                break
+            with suppress(TimeoutExpired):
+                adb_result = run(
+                    cmd,
+                    capture_output=True,
+                    check=False,
+                    timeout=30,
+                )
+                if adb_result.returncode == 0 and adb_result.stdout.strip() == b"1":
+                    LOG.debug("device booted")
+                    break
             if proc.poll() is not None:
                 raise AndroidEmulatorError("Failed to launch emulator.")
             if perf_counter() >= deadline:
@@ -629,7 +625,6 @@ class AndroidEmulator:
         Returns:
             None
         """
-        assert self.emu is not None
         self.emu.terminate()
 
     def poll(self) -> int | None:
@@ -641,10 +636,9 @@ class AndroidEmulator:
         Returns:
             Exit status of emulator process (None if still running).
         """
-        assert self.emu is not None
         return self.emu.poll()
 
-    def wait(self, timeout: int | None = None) -> int | None:
+    def wait(self, timeout: float | None = None) -> int:
         """Wait for emulator process to exit.
 
         Args:
@@ -652,9 +646,8 @@ class AndroidEmulator:
                      subprocess.TimeoutExpired.
 
         Returns:
-            Exit status of emulator process (None if still running).
+            Exit status of emulator process.
         """
-        assert self.emu is not None
         return self.emu.wait(timeout=timeout)
 
     def save_snapshot(self) -> None:
@@ -687,7 +680,6 @@ class AndroidEmulator:
         Returns:
             None
         """
-        LOG.info("Initiating emulator shutdown")
         # terminate is handled and the emulator attempts a clean shutdown
         self.terminate()
 
@@ -750,9 +742,9 @@ class AndroidEmulator:
         """
         mksd_path = PATHS.sdk_root / "emulator" / f"mksdcard{EXE_SUFFIX}"
         assert mksd_path.is_file(), f"Missing {mksd_path}"
-        LOG.info("Creating AVD '%s'", avd_name)
 
         # create an avd
+        LOG.debug("creating AVD '%s'", avd_name)
         PATHS.avd_home.mkdir(exist_ok=True)
         api_gapi = PATHS.sdk_root / "system-images" / SYS_IMG / "default"
         cls.remove_avd(avd_name)
@@ -836,6 +828,9 @@ def main(argv: list[str] | None = None) -> None:
     aparser.add_argument(
         "--verbose", "-v", action="store_true", help="Enable verbose logging"
     )
+    aparser.add_argument(
+        "--emulator-output", action="store_true", help="Display emulator console output"
+    )
     aparser.add_argument(
         "--boot-timeout",
         "-t",
@@ -876,16 +871,21 @@ def main(argv: list[str] | None = None) -> None:
     init_logging(debug=args.verbose)
 
     if not args.skip_dl:
-        AndroidEmulator.install()
+        try:
+            AndroidEmulator.install()
+        except KeyboardInterrupt:
+            LOG.info("Aborting...")
+            return
 
     if not args.no_launch:
         # Find a free port
         port = AndroidEmulator.search_free_ports()
         avd_name = f"x86.{port:d}"
-
-        # Create an AVD and boot it once
-        AndroidEmulator.create_avd(avd_name)
+        emu: AndroidEmulator | None = None
         try:
+            # Create an AVD and boot it once
+            AndroidEmulator.create_avd(avd_name)
+            LOG.info("Launching Android emulator...")
             # Boot the AVD
             emu = AndroidEmulator(
                 port=port,
@@ -895,20 +895,20 @@ def main(argv: list[str] | None = None) -> None:
                 headless=args.headless,
                 xvfb=args.xvfb,
             )
-            LOG.info("Android emulator is running on port %d", port)
-            try:
-                emu.wait()
-            finally:
-                if emu.poll() is None:
-                    emu.shutdown()
-                try:
-                    # this should never timeout
-                    emu.wait(timeout=120)
-                finally:
-                    emu.cleanup()
+            LOG.info("Android emulator is running as 'emulator-%d'", port)
+            emu.wait()
         except KeyboardInterrupt:
             LOG.info("Aborting...")
         finally:
+            LOG.info("Initiating emulator shutdown")
+            if emu is not None:
+                if emu.poll() is None:
+                    emu.terminate()
+                try:
+                    # this should never timeout
+                    emu.wait(timeout=60)
+                finally:
+                    emu.cleanup()
             AndroidEmulator.remove_avd(avd_name)
 
 
